#include <assert.h>
#include "engine.h"
#include "main.h"
#include "color.h"
#include "screenshot.h"
#include "background.h"
#include "player.h"
#include "leveldata.h"

int Engine::getCounter ()
{
	return parent->getCounter();
}

Engine::Engine (Main *p) : layers (), menubg(*this), menu (*this), floor (*this), sprites (*this),
	view(), playerControl (*this), script (*this), insults (*this)
{
	parent = p;
	debug = false;
	resources = p->getResources();
	settings = p->getSettings();
//	dp = NULL;

	layers.push_back (&menubg);
	layers.push_back (&menu);

	view.setOfst (0, 0);

	floor.setViewPort (&view);
	sprites.setViewPort (&view);
	playerControl.setViewPort (&view);

	layers.push_back (&floor);
	layers.push_back (&sprites);
	layers.push_back (&playerControl);
	layers.push_back (&script);
	layers.push_back (&insults);

	btnScreenshot.setScancode (KEY_F12);

	btnQuit.setScancode (KEY_ESC);
#ifdef DEBUG
	btnDebugMode.setScancode (KEY_F11);
#endif
}

Button* Engine::getButton()
{
	return settings->getButton();
}

void Engine::init()
{
 	srand(time(0));

 	ActiveSprite::init (resources);

 	menubg.init();
 	menu.init();
 	floor.init();
 	insults.init();
 	script.init();

 	startMenu();

}

void Engine::startMenu()
{
 	menubg.background();
 	menu.foreground();

	sprites.stop();
	playerControl.stop();
	floor.stop();
}

void Engine::stopMenu()
{
 	menu.stop();
 	menubg.stop();
}

void Engine::startGame()
{
 	playerControl.initGame();
 	startScript();
}

void Engine::startScript()
{
	clear_keybuf();
 	script.prepare(playerControl.getCurrentLevel());
	script.foreground();
}

void Engine::stopScript()
{
	script.stop();
}

void Engine::startLevel()
{
 	initLevel(playerControl.getCurrentLevel());
 	floor.background();
 	sprites.background();
 	playerControl.foreground();
 	insults.foreground();
}

void Engine::stopLevel()
{
 	floor.stop();
 	sprites.stop();
 	playerControl.stop();
 	insults.stop();
}

void Engine::shutdown()
{
	parent->stop();
}

void Engine::draw (BITMAP *buffer)
{
	for (list<Layer*>::iterator i = layers.begin(); i != layers.end(); ++i)
	{
		if ((*i)->getState() != Layer::LAYER_SLEEP)
			(*i)->draw(buffer);
	}
}

void Engine::update ()
{
	if (btnScreenshot.justPressed())
	{
		screenshot();
	}

#ifdef DEBUG
	if (btnDebugMode.justPressed())
	{
		debug = !debug;
	}
#endif
	if (btnQuit.justPressed())
	{
		shutdown();
	}


	for (list<Layer*>::iterator i = layers.begin(); i != layers.end(); ++i)
	{
		Layer::LayerState state = (*i)->getState();

		if (state == Layer::LAYER_FOREGROUND)
			(*i)->handleInput();

		if (state != Layer::LAYER_SLEEP)
			(*i)->update();
	}
}

void Engine::playSample (SAMPLE *s)
{
	if (!settings->isSoundOn()) return;
	play_sample (s, 127, 127, 1000, 0);
}

void Engine::playMusic (const char *id, float volume)
{
/*
    if (!(settings->isSoundOn() && settings->isMusicOn())) return;
    if (dp)
    {
        al_stop_duh (dp);
        dp = NULL;
    }
    dp = al_start_duh (resources->getDuh (id), 2, 0, volume, 4096, 22050);
*/
}

void Engine::stopMusic ()
{
/*
    if (dp)
    {
        al_stop_duh (dp);
        dp = NULL;
    }
*/
}

void Engine::done()
{
	stopMusic();
}

int Engine::getMaxZOrder()
{
	int max = 0;
	for (list<Layer*>::iterator i = layers.begin(); i != layers.end(); ++i)
	{
		if ((*i)->getZOrder() > max)
		{
			max = (*i)->getZOrder();
		}
	}
	return max;
}

void Engine::initLevel(int level)
{
	LevelData const * current = &levels[level];

	sprites.killAll(); // clean from previous game

	// add sheds and buildings
	for (int i = 0; i < current->buildings; i++)
	{
		SpriteSubType st = (SpriteSubType)(rand() % 5);
		int x = rand() % current->map_sizex;
		int y = rand() % current->map_sizey;

		Solid *s = new Solid (this, st);
		s->setLocation (x, y, 0);
		sprites.add (s);
	}

	// add shrubs and bushes


	// add chopper && tank
	playerControl.initLevel(current->haschopper, current->hastank, current->targets);
	
	// add turrets
	for (int i = 0; i < current->enemy_turrets; ++i)
	{
		Enemy *e = new Enemy(this, SST_TURRET);
		int x, y;
		do
		{
			x = rand() % current->map_sizex;
			y = rand() % current->map_sizey;
		} while (x < 150.0 && y < 150.0);
		e->setLocation (x, y, 0);
		sprites.add (e);
	}

	// add SAM sites
	for (int i = 0; i < current->enemy_sam; ++i)
	{
		Enemy *e = new Enemy(this, SST_SAM);
		int x, y;
		do
		{
			x = rand() % current->map_sizex;
			y = rand() % current->map_sizey;
		} while (x < 300.0 && y < 300.0);
		e->setLocation (x, y, 0);
		sprites.add (e);
	}

	// add enemy tanks
	for (int i = 0; i < current->enemy_tanks; ++i)
	{
		Enemy *e = new Enemy(this, SST_ENEMYTANK);
		int x, y;
		do
		{
			x = rand() % current->map_sizex;
			y = rand() % current->map_sizey;
		} while (x < 200.0 && y < 200.0);
		e->setLocation (x, y, 0);
		sprites.add (e);
	}

	// add enemy bunkers
	for (int i = 0; i < current->enemy_bunker; ++i)
	{
		Enemy *e = new Enemy(this, SST_BUNKER);
		int x, y;
		do
		{
			x = rand() % current->map_sizex;
			y = rand() % current->map_sizey;
		} while (x < 200.0 && y < 200.0);
		e->setLocation (x, y, 0);
		sprites.add (e);
	}

	// add goal bunker
	for (int i = 0; i < current->targets; ++i)
	{
		Enemy *e = new Enemy(this, SST_MAINBUNKER);
		int x, y;
		do
		{
			x = rand() % current->map_sizex;
			y = rand() % current->map_sizey;
		} while (x < 800.0 && y < 800.0);
		e->setLocation (x, y, 0);
		sprites.add (e);
	}
}
