#include "player.h"
#include "color.h"
#include "engine.h"
#include "isometric.h"
#include <math.h>
	
ActiveSprite::ActiveSprite (Engine *e, SpriteSubType _st) : Sprite (e), speed (0), current (NULL), target(NULL), subType (_st), health (100)
{
	nozzlex = 0;
	nozzley = 0;
	nozzlez = 0;
	
	SpriteData *data = &spriteData[_st];
	// solid = data->solid;
	shadow = data->shadow;
	
	if (data->anim != NULL)
	{
		current = data->anim;
		sizex = current->sizex;
		sizey = current->sizey;
		sizez = current->sizez;
	}	
}

/**
	Set target and adjust locking of target
	param may be NULL (stops targeting)
*/
void ActiveSprite::setTarget(Sprite *_target)
{
	if (target != _target)
	{
		if (_target == NULL)
		{
			target->lock--;
		}
		target = _target;
		if (target != NULL)
		{
			target->lock++;
		}
	}
}

void Bullet::setSource (Sprite *_source)
{
	if (source != _source)
	{
		if (_source == NULL)
		{
			source->lock--;
		}
		source = _source;
		if (source != NULL)
		{
			source->lock++;
		}
	}
}

Bullet::Bullet(Engine *e, Sprite *_source, Sprite *_target, Direction startDir, float startSpeed) : ActiveSprite (e, SST_MISSILE) 
{
	source = NULL;
	setSource (_source);
	tLive = 0;
	solid = true;
	color = RED;
	dx = 0;
	dy = 0;
	dz = 0;
	setTarget (_target);
	angle = (startDir * M_PI / 4);
	speed = startSpeed;
	maxLive = MISSILE_LIVE_TIME;
	damage = 8;
}

Bullet::Bullet(Engine *e, Sprite *_source, float _dx, float _dy, float _dz) : ActiveSprite (e, SST_BULLET) 
{
	source = NULL;
	setSource (_source);
	tLive = 0;
	solid = true;
	color = WHITE;
	dx = _dx;
	dy = _dy;
	dz = _dz;
	setTarget (NULL);
	angle = 0;
	speed = 0;
	maxLive = LIVE_TIME;
	damage = 3;
}

void Bullet::update()
{
	switch (getSubType())
	{
		case SST_BULLET:
				try_move (dx, dy, dz);
			break;
		case SST_MISSILE:
			{
				float dest, vector_angle;
				delta_vector(this, target, dest, vector_angle);
				
				float delta_angle = angle - vector_angle + M_PI;
				normalize_angle (delta_angle);
				if (delta_angle < -0.1) { angle -= 0.05; }
				if (delta_angle > 0.1) { angle += 0.05; }
				
				float ddz = getz() - target->getz();
				
				if (ddz < -1.0)
				{
					dz = 1.0;
				}
				if (ddz > 1.0 && getz() > 5.0)
				{
					dz = -1.0;
				}
				
				if (speed < 20.0) speed += 0.3; 
				
				dx = speed * cos (angle);
				dy = speed * sin (angle);
				
				//TODO: homing
				try_move (dx, dy, dz);
			}
			break;
	}
	if (++tLive > maxLive || getz() == 0) kill();
}

void Bullet::handleCollission(Sprite * s)
{
	if (s == source) { return; } // can't hit yourself
	if (s->getType() != source->getType()) // enemy <-> player or vice versa
	{
		kill();
		if (getSubType() == SST_BULLET)
		{
			explode (0);
		}
		else
		{	
			explode (1);
		}
	}
}

Explosion::Explosion(Engine *e, int size) : Sprite (e) 
{
	solid = false;
	color = YELLOW;
	maxLive = size;
	sizex = 1;
	sizey = 1;
	sizez = 1;
	tLive = 0;
}

void Explosion::draw(BITMAP *buffer, int xofst, int yofst)
{
	int rx, ry;
	isoToScreen (getx(), gety(), getz(), rx, ry);
	
	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	set_trans_blender (0, 0, 0, 128);
	
	circlefill (buffer, rx - xofst, ry - yofst, sizex, RED);
	circlefill (buffer, rx - xofst, ry - yofst, sizex * 2 / 3, YELLOW);
	
	drawing_mode (DRAW_MODE_SOLID, NULL, 0, 0);	
}

void Explosion::update()
{
	sizex++;
	if (++tLive > maxLive)
	{
		kill();
	}
}

Player::Player(Engine *e, SpriteSubType st) : ActiveSprite (e, st) 
{
	accelerate = false;
	solid = true;
	color = GREEN;
}

Enemy::Enemy(Engine *e, SpriteSubType st) : ActiveSprite (e, st) 
{
	tReload = 0;
	solid = true;
	color = BLUE;
	
	target = NULL;
	switch (st)
	{
		case SST_ENEMYTANK:
			viewRange = 500;
			mobile = true;
			nozzlez = 5;
			break;
		case SST_TURRET:
			viewRange = 200;
			mobile = false;
			nozzlez = 5;
			break;
		case SST_BUNKER:
			viewRange = 200;
			mobile = false;
			nozzlez = 5;
			break;
		case SST_SAM:
			viewRange = 200;
			mobile = false;
			nozzlez = 100;
			break;
		case SST_MAINBUNKER:
			viewRange = 200;
			mobile = false;
			nozzlez = 5;
			break;
		default:
			assert (false);
			break;
	}
}

void Enemy::update()
{
	if (!getTarget())
	{
		if (getSubType() == SST_SAM)
		{
			setTarget (engine->getPlayerControl()->getChopper());
		}
		else
		{
			setTarget (engine->getPlayerControl()->getNearestPlayer(getx(), gety(), viewRange));
		}
	}
	
	if (getTarget())
	{
		float angle, dist;
		delta_vector(this, target, dist, angle);
		
		setDir (dirFromRadians(angle));
		if (getSubType() == SST_ENEMYTANK)
		{
			if (dist > (200.0 * 200.0))
			{
				speed += 0.2;
				if (speed > 4.0) speed = 4.0;
			}
		}
		
		if (dist > (viewRange * viewRange))
			setTarget (NULL);
			
		
	}
	
	if (getSubType() == SST_ENEMYTANK)
	{
		if (speed < 0.1) speed = 0;
		float angle = radiansFromDir(getDir());
		float dx = speed * cos(angle);
		float dy = speed * sin(angle);
		try_move (dx, dy, 0);
		speed -= 0.1; // decelerate
	}
	
	// convert back to round
	if (++tReload > RELOAD_TIME)
	{
		tReload = 0;
		float radians = radiansFromDir (getDir());
		
		if (getTarget())
		{
			switch (getSubType())
			{
			case SST_SAM:
				fireMissile();
				break;
			case SST_TANK:
			case SST_TURRET:
				fireBullet4();
				break;			
			default: // bunkers etc.
				fireBullet2 ();
				break;
			}
		}
	}
}

void Enemy::handleCollission(Sprite * s)
{
	if (s->getType() == ST_BULLET)
	{
		Bullet *b = dynamic_cast<Bullet *>(s);
		assert (b); // must be a bullet
		if (b->getSource()->getType() == ST_ENEMY)
		{
			return; // ignore self-hit
		}
		health -= 10;
		if (health < 0)
		{
			health = 0;
			kill();
			if (subType == SST_MAINBUNKER)
			{
				engine->getPlayerControl()->decreaseTargets();
				explode(5);
			}
			else
			{
				explode(2);
			}
		}
	}
}


Bonus::Bonus(Engine *e, SpriteSubType st) : ActiveSprite (e, st) 
{
	solid = true;
	color = WHITE;
}

Solid::Solid(Engine *e, SpriteSubType st) : ActiveSprite (e, st) 
{
	switch (st)
	{
	case SST_RUBBLE:
		solid = false;
		color = GREY;
		break;
	case SST_BUSH:
		solid = false;
		color = DARK_GREEN;
		break;
	case SST_HOUSE:
		solid = true;
		color = GREY;
		break;
	case SST_TREE:
		solid = true;
		color = DARK_GREEN;
		break;
	case SST_TALL:
		solid = true;
		color = GREY;
		break;
	default:
		assert (false);
		break;
	}
}

// chopper
void ActiveSprite::fireBullet()
{
	float radians = radiansFromDir(getDir());
	float bulletSpeed = speed + 6.0;
	Bullet *b = new Bullet(engine, this, bulletSpeed * cos(radians), bulletSpeed * sin(radians), -(getz() / 50));
	b->setLocation(getx(), gety(), getz() + 5.0);
	parent->add(b);
}

// player tank
void ActiveSprite::fireBullet3()
{
	float radians = radiansFromDir(getDir());
	float bulletSpeed = speed + 6.0;
	Bullet *b = new Bullet(engine, this, bulletSpeed * cos(radians), bulletSpeed * sin(radians), 0);
	b->setLocation(getx(), gety(), getz() + 5.0);
	parent->add(b);
}

// enemy bunker
void ActiveSprite::fireBullet2()
{
	float dx = target->getx() - getx();
	float dy = target->gety() - gety();
	float dz = target->getz() - getz();
	float dist = sqrt ((dx * dx) + (dy * dy));
	
	if (dist == 0) return; // div by 0 prevention
	
	float bulletSpeed = 6.0;
	Bullet *b = new Bullet(engine, this, dx / dist * bulletSpeed, dy / dist * bulletSpeed, 0);
	
	b->setLocation(getx(), gety(), getz() + nozzlez);
	parent->add(b);
}

// enemy turret
void ActiveSprite::fireBullet4()
{
	float dx = target->getx() - getx();
	float dy = target->gety() - gety();
	float dz = target->getz() - getz();
	float dist = sqrt ((dx * dx) + (dy * dy) + (dz * dz));
	
	if (dist == 0) return; // div by 0 prevention
	
	float bulletSpeed = 6.0;
	Bullet *b = new Bullet(engine, this, dx / dist * bulletSpeed, dy / dist * bulletSpeed, dz / dist * bulletSpeed);
	
	b->setLocation(getx(), gety(), getz() + 100.0);
	parent->add(b);
}

void ActiveSprite::fireMissile()
{
	Bullet *b = new Bullet(engine, this, getTarget(), getDir(), speed);
	b->setLocation(getx(), gety(), getz() + 100);
	parent->add(b);
}

void ActiveSprite::explode(int scale)
{
	bool tremble = false;
	float ampl = 0.0;
	int size = 5;
	
	switch (scale)
	{
		case 0: // bullet impact
			tremble = false;
			size = 5;
			break;
		case 1: // missile impact
			tremble = false;
			size = 10;
			break;
		case 2: // building destroyed
			tremble = true;
			ampl = 10.0;
			size = 50;
			break;
		case 3: // player unit destroyed
			tremble = true;
			ampl = 3.0;
			size = 30;
			break;
		case 5: // mainbunker destroyed
			tremble = true;
			size = 70;
			ampl = 20.0;
			break;
	}
	
	Explosion *e = new Explosion (engine, size);
	e->setLocation(getx(), gety(), getz());
	parent->add(e);
	if (tremble)
	{
		parent->getViewPort()->tremble(ampl);
	}
}

void ActiveSprite::init(Resources *res)
{
	for (int i = 0; i < SPRITE_NUM; ++i)
	{
		if (spriteData[i].animName != NULL)
		{
			spriteData[i].anim = res->getAnim (spriteData[i].animName);
		}
	}
}

void ActiveSprite::draw(BITMAP *buffer, int xofst, int yofst)
{
	if (current)
	{
		int rx, ry;
		isoToScreen(getx(), gety(), getz(), rx, ry);
		current->drawFrame (buffer, 0, getDir(), rx - xofst, ry - yofst);
		
		if (engine->isDebug())
		{
			Sprite *t = getTarget();
			if (t)
			{
				int r2x, r2y;
				isoToScreen(t->getx(), t->gety(), t->getz(), r2x, r2y);
				line (buffer, rx - xofst, ry - yofst, r2x - xofst, r2y - yofst, YELLOW);
			}
		}
	}
	else
	{
		Sprite::draw (buffer, xofst, yofst);
	}
}

void Player::setAccelerate()
{
	accelerate = true;
}

void Player::update()
{	
	if (accelerate)
	{
		speed += PLAYER_SPEED_INC;
		if (speed > PLAYER_MAX_SPEED) speed = PLAYER_MAX_SPEED;
	}
	else
	{
		speed -= PLAYER_SPEED_INC;
		if (speed < 0) speed = 0;
	}
	accelerate = false; // needs to be activated constantly
	
	float radians = radiansFromDir(getDir());
	
	float dx = speed * cos (radians);
	float dy = speed * sin (radians);

	try_move (dx, dy, 0);
}

void Player::handleCollission(Sprite * s)
{
	if (s->getType() == ST_BULLET)
	{
		Bullet *b = dynamic_cast<Bullet *>(s);
		assert (b); // must be a bullet
		if (b->getSource() == this)
		{
			return; // ignore self-hit
		}
		health -= b->getDamage();
		if (health < 0)
		{
			health = 0;
			engine->getPlayerControl()->playerKilled(this);
			kill();
			explode(3);
		}
	}
}

float sq_dist (Sprite *src, Sprite *dest)
{
	float dx = dest->getx() - src->getx();
	float dy = dest->gety() - src->gety();
	return (dx * dx) + (dy * dy);
}

void delta_vector (Sprite *src, Sprite *dest, float &sq_dist, float &radians)
{
	float dx = dest->getx() - src->getx();
	float dy = dest->gety() - src->gety();
	sq_dist = (dx * dx) + (dy * dy);
	radians = atan2 (dy, dx);
}

void normalize_angle (float &angle)
{
	while (angle < -M_PI) { angle += 2 * M_PI; }
	while (angle > M_PI) { angle -= 2 * M_PI; }
}
