#include "playercontrol.h"
#include "player.h"
#include "button.h"
#include <assert.h>
#include "engine.h"
#include "color.h"
#include "math.h"
#include "leveldata.h"
#include "isometric.h"

PlayerControl::PlayerControl(Engine &e) : Layer (e)
{
	btnCheat.setScancode(KEY_F9);
}

void PlayerControl::initGame()
{
	score = 0;
	lives = 5;
	currentLevel = 0;
	chopper = NULL;
	tank = NULL;
}

void PlayerControl::initLevel(bool hasChopper, bool hasTank, int _targetsRemaining)
{
	current = NULL;
	if (hasTank) initTank();
	if (hasChopper) initChopper();
	targetsRemaining = _targetsRemaining;
	tOutro = 0;
	outro = false;
	nextTarget = NULL;
	assert (viewPort);
	viewPort->setOfst(0, 0);
}

void PlayerControl::initTank()
{
	tank = new Player(&parent, SST_TANK);
	tank->setLocation (20, 10, 0);
	parent.getSpriteLayer()->add (tank);
	if (current == NULL) current = tank;
}

void PlayerControl::initChopper()
{
	// add player chopper
	chopper = new Player(&parent, SST_CHOPPER);
	chopper->setLocation (10, 50, 10);
	parent.getSpriteLayer()->add (chopper);
	if (current == NULL) current = chopper;
}

void PlayerControl::draw(BITMAP *buffer)
{
	textprintf_right_ex(buffer, font, buffer->w, 0, BLACK, -1, "HEALTH %3i LIVES %02i REMAIN %i", 
		current->getHealth(), lives, targetsRemaining);

	if (parent.isDebug())
	{
		textprintf_ex(buffer, font, 0, 16, DARK_GREEN, -1, "X: %f Y: %f Z: %f",
			current->getx(), current->gety(), current->getz());
	}
	
	if (nextTarget)
	{
		float dx = - current->getx() + nextTarget->getx();
		float dy = - current->gety() + nextTarget->gety();
		
		float dist = dx * dx + dy * dy;
		if (dist > 500 * 500)
		{
			float angle = atan2 (dy, dx);

			float rdx, rdy;
			isoToScreen_f (cos (angle), sin(angle), 0, rdx, rdy);
			
			int cx = buffer->w / 2;
			int cy = buffer->h / 2;
			
			const float ARROW = 200;
			int points[6] =
			{
				cx + ARROW * rdx,
				cy + ARROW * rdy,
				cx + (ARROW - 20) * rdx - 10 * rdy,
				cy + (ARROW - 20) * rdy + 10 * rdx,
				cx + (ARROW - 20) * rdx + 10 * rdy,
				cy + (ARROW - 20) * rdy - 10 * rdx,				 
			};
			
			polygon (buffer, 3, points, GREEN); 
			
			//line (buffer, cx, cy, cx + rdx * 50, cy + rdy * 50, GREEN);
		}
	}
}

void PlayerControl::handleInput()
{
	assert (current);
	
#ifdef DEBUG
	if (btnCheat.justPressed())
	{
		nextLevel();
	}
#endif
	// handle player input
	Button *b = parent.getButton();
	
	if (b[btnSwitch].justPressed())
	{
		Player *other = (current == tank) ? chopper : tank;
		if (other != NULL)
			current = other;
	}
		
	Direction newDir = current->getDir();
	bool keyDown = false;
	
	if (b[btnN].getState()) { newDir = N; keyDown = true; }
	if (b[btnS].getState()) { newDir = S; keyDown = true; }
	if (b[btnE].getState()) { newDir = E; keyDown = true; }
	if (b[btnW].getState()) { newDir = W; keyDown = true; }
	if (b[btnN].getState() && b[btnE].getState()) { newDir = NE; keyDown = true; }
	if (b[btnN].getState() && b[btnW].getState()) { newDir = NW; keyDown = true; }
	if (b[btnS].getState() && b[btnE].getState()) { newDir = SE; keyDown = true; }
	if (b[btnS].getState() && b[btnW].getState()) { newDir = SW; keyDown = true; }
	if (b[btnNE].getState()) { newDir = NE; keyDown = true; }
	if (b[btnSE].getState()) { newDir = SE; keyDown = true; }
	if (b[btnNW].getState()) { newDir = NW; keyDown = true; }
	if (b[btnSW].getState()) { newDir = SW; keyDown = true; }
	
	if (keyDown)
	{
		current->setDir (newDir);
		current->setAccelerate();
	}
		
	// fly up / down with chopper
	if (current == chopper)
	{
		float dz = 0;
		if (current->getz() < MAXZ && b[btnUp].getState())
		{
			dz = 1;
		}
		if (b[btnDown].getState())
		{
			dz = -1;
		}
	
		current->try_move(0, 0, dz);
	}
	
	if (b[btnFire].justPressed())
	{
		if (current == chopper)
		{
			current->fireBullet();
		}
		else
		{
			current->fireBullet3();
		}
	}
	
	if (b[btnAltFire].justPressed())
	{
		//TODO: find way to set target
		//current->fireMissile();
	}
	
	assert (viewPort);
	
	float radians = radiansFromDir(current->getDir());
	float goalx = current->getx() + (current->speed * 30 * cos (radians));
	float goaly = current->gety() + (current->speed * 30 * sin (radians));
	
	int rx, ry;
	isoToScreen(goalx, goaly, 0, rx, ry);

	viewPort->moveTo (rx - SCREEN_W / 2, ry - SCREEN_H / 2);
	viewPort->update();
}

Player* PlayerControl::getNearestPlayer(float x, float y, float maxrange)
{
	Player *min = current;
	float xx = min->getx();
	float yy = min->gety();
	float mindist = (x - xx) * (x - xx) + (y - yy) * (y - yy);
	
	if (tank)
	{
		xx = tank->getx();
		yy = tank->gety();
		float dist = (x - xx) * (x - xx) + (y - yy) * (y - yy);
		if (dist < mindist)
		{
			dist = mindist;
			min = tank;
		}
	}
	
	if (chopper)
	{
		xx = chopper->getx();
		yy = chopper->gety();
		float dist = (x - xx) * (x - xx) + (y - yy) * (y - yy);
		if (dist < mindist)
		{
			dist = mindist;
			min = chopper;
		}
	}
	
	if (mindist > (maxrange * maxrange))
	{
		return NULL;
	}
	else
	{
		return min;
	}
}

void PlayerControl::update()
{
	if (!nextTarget)
	{
		nextTarget = parent.getSpriteLayer()->getNextBunker ();
	}

	if (outro)
	{
		if (++tOutro > OUTRO_TIME)
		{
			nextLevel();
		}
	}
}

void PlayerControl::decreaseTargets()
{
	targetsRemaining--;
	nextTarget = NULL;
	if (targetsRemaining == 0)
	{
		outro = true;
	}
}

void PlayerControl::decreaseLife()
{
	lives--;
	if (lives == 0)
	{
		gameOver();
	}
}

void PlayerControl::gameOver()
{
	currentLevel = LEVEL_GAMEOVER;
	parent.stopLevel();
	parent.startScript();
}

void PlayerControl::nextLevel()
{
	currentLevel++;
	if (currentLevel == LEVEL_NUM)
	{
		currentLevel = LEVEL_WON;
	}
	parent.stopLevel();
	parent.startScript();
}

void PlayerControl::playerKilled(Player * n)
{
	decreaseLife(); // goes gameover if life == 0
	
	if (n == current) current = NULL;
	
	if (n == chopper)
	{
		chopper = NULL;
		initChopper();
	}
	if (n == tank)
	{
		tank = NULL;
		initTank();
	}
}
