#include "script.h"
#include <assert.h>
#include "engine.h"
#include "color.h"
#include <fstream>
#include "leveldata.h"
#include <iostream>

Script::Script(Engine &e) : Layer (e), scriptmap(), ip(), lines()
{
	state = SCRIPT_STOP; // call prepare() to start
}

void Script::prepare (int _level)
{
	level = _level;
	if (level == LEVEL_GAMEOVER)
	{
		currentScript = scriptmap["gameover"];
	}
	else if (level == LEVEL_WON)
	{
		currentScript = scriptmap["won"];
	}
	else
	{
		char buffer[100];
		snprintf (buffer, sizeof (buffer), "level%i", level + 1);
		currentScript = scriptmap[buffer];
	}
	ip = currentScript.begin();
	
	state = SCRIPT_PAUSE;
	
	tPause = 0;
	pause_time = 30;
	lines.clear();
	textx = 20;
	texty = 20;
	textw = SCREEN_W;
	texth = SCREEN_H;
	imgx = 0;
	imgy = 0;
	img = NULL;
}

void Script::runScript ()
{
	if (ip == currentScript.end())
	{
		state = SCRIPT_STOP;
		parent.stopScript();
		if (level == LEVEL_WON || level == LEVEL_GAMEOVER)
		{
			parent.startMenu();
		}
		else
		{
			parent.startLevel();
		}
		return;
	}
	switch (ip->cmd)
	{
		case CMD_SAY:
			lines.push_back (ip->args);
			break;
		case CMD_WAIT:
			state = SCRIPT_PAUSE;
			pause_time = ip->arg1;
			break;
		case CMD_PAUSE:
			clear_keybuf();
			state = SCRIPT_WAIT_KEYPRESS;
			break;
		case CMD_IMG_CENTER:
			img = parent.getResources()->getRle(ip->args);
			imgx = (SCREEN_W - img->w) / 2;
			imgy = (SCREEN_H - img->h) / 2;
			break;
		case CMD_IMG:
			img = parent.getResources()->getRle(ip->args);
			imgx = ip->arg1;
			imgy = ip->arg2;
			break;
	}
	ip++;
}

void Script::draw(BITMAP *buffer)
{	
	clear_to_color (buffer,  BLACK);
	
	vector<string>::iterator i;
	
	int x = textx;
	int y = texty;
	
	for (i = lines.begin(); i != lines.end(); ++i)
	{
		textout_ex (buffer, font, i->c_str(), x, y, WHITE, -1);
		y += LINEH;
	}
	
	if (img)
	{
		draw_rle_sprite (buffer, img, imgx, imgy);
	}
}

void Script::update()
{
	switch (state)
	{
		case SCRIPT_RUN:
			while (state == SCRIPT_RUN) { runScript(); }
			break;
		case SCRIPT_WAIT_KEYPRESS:
			if (keypressed())
			{
				readkey();
				lines.clear(); // clear current text box
				state = SCRIPT_RUN;
			}
			break;
		case SCRIPT_PAUSE:
			tPause++;
			if (tPause > pause_time)
			{
				tPause = 0;
				pause_time = 0;
				state = SCRIPT_RUN;
			}
			break;
		case SCRIPT_STOP:
			// script has ended, do nothing.
			break;
	}
}

void Script::readFromFile(char * filename)
{
	char errMsg[100];
	bool error = false;
	
	vector<Instruction> script;
	
	ifstream infile (filename, ios::in);
	if (!infile)
	{
		snprintf (errMsg, sizeof(error), "Could not open file '%s'", filename);
		error = true;
	}
	
	char temp [1024];
	char line [1024];
	
	string section = "main"; // default section
	
	Instruction i;
	while (!error && !infile.eof())
	{
		infile.getline (line, sizeof (line));
		
		if (line[0] == '[')
		{
			// start of section
			char *start = strchr(line, '[') + 1; 
			char *end = strchr(line, ']');
			
			assert (start);
			assert (end);
			
			int len = end - start;
			strncpy(temp, start, len);
			temp[len] = '\0';
			 
			scriptmap.insert (pair<string, vector<Instruction> >(section, script));
			script.clear();
			section = string (temp);
		}
		else if (strncmp (line, "SAY", 3) == 0)
		{
			char *start = strchr(line, ' ') + 1; 
			char *end = strchr(line, '\0');
			
			assert (start);
			assert (end);
			
			int len = end - start + 1;
			strncpy(temp, start, len);
			
			i.cmd = CMD_SAY;
			i.args = string (temp);
			script.push_back(i);
		}
		else if (strncmp (line, "CENTER", 6) == 0)
		{
			char *start = strchr(line, ' ') + 1; 
			char *end = strchr(line, '\0');
			
			assert (start);
			assert (end);
			
			int len = end - start + 1;
			strncpy(temp, start, len);
			
			i.cmd = CMD_IMG_CENTER;
			i.args = string (temp);
			script.push_back(i);
		}
		else if (strncmp (line, "IMAGE", 5) == 0)
		{
			int code = sscanf (line, "IMAGE %i %i %s", &i.arg1, &i.arg2, temp);
			assert (code == 3);
			i.args = string (temp);
			i.cmd = CMD_IMG;
			script.push_back(i);
		}
		else if (strncmp (line, "PAUSE", 5) == 0)
		{
			i.cmd = CMD_PAUSE;
			script.push_back(i);
		}
		else if (strcmp (line, "") == 0)
		{
			// ignore emtpy line
		}
		else
		{
			error = true;
			snprintf (errMsg, sizeof (errMsg), "Could not parse line '%s'", line);
		}
	}
	
	if (!error)
	{
		scriptmap.insert (pair<string, vector<Instruction> >(section, script));
		script.clear();
	}
	
	if (error)
	{
		allegro_message (errMsg);
		assert (false);
	}
/*	
// print script
	for (map<string, vector<Instruction> >::iterator i = scriptmap.begin(); i != scriptmap.end(); ++i)
	{
		cout << i->first << endl;
		for (vector<Instruction>::iterator j = i->second.begin(); j != i->second.end(); ++j)
		{
			cout << "  " << j->cmd << " " << j->args << endl;
		}
	}
*/
}

void Script::init()
{
	readFromFile ("data/scripts.txt");
}
