#include <assert.h>
#include "sprite.h"
#include "engine.h"
#include "color.h"
#include <math.h>
#include "isometric.h"

using namespace std;

void Sprite::setDir(Direction _dir) 
{ 
	if (_dir != dir) 
	{ 
		dir = _dir; 
	} 
}

Sprite::Sprite (Engine *e) : engine (e)
{
	lock = 0;
	x = 0;
	y = 0;
	z = 0;
	sizex = 50;
	sizey = 50;
	sizez = 50;
	color = RED;
	alive = true;
	visible = true;
	solid = true;
	dir = E;
	parent = NULL;
	shadow = false;
}

void Sprite::setLocation (float nx, float ny, float nz)
{
	x = nx;
	y = ny;
	z = nz;
}

void Sprite::setParent (SpriteLayer *_parent)
{
	parent = _parent;
}

void SpriteLayer::killAll()
{
	list<Sprite*>::iterator i;
	for (i = sprites.begin(); i != sprites.end(); ++i)
	{
		delete (*i);
		(*i) = NULL;
	}
	sprites.clear();
}

void Sprite::update()
{
	assert (parent);
}

void SpriteLayer::add(Sprite *o)
{
	sprites.push_back (o);
	o->setParent (this);
}

class MyObjectRemover
{
   public:
	  bool operator()(Sprite *s)
	  {
		 if (s->canDestroy())
		 {
			delete s;
			return 1;
		 }
		 return 0;
	  }
};

void SpriteLayer::update()
{
	testSprites();
	
	list<Sprite*>::iterator i;
	for (i = sprites.begin(); i != sprites.end(); i++)
	{
		if ((*i)->isAlive()) (*i)->update();
	}
	
	testSprites();

	// collission detection!	
	list<Sprite*>::iterator j;
	for (i = sprites.begin(); i != sprites.end(); i++)
		for (j = sprites.begin(); j != i; j++)
	{
		// see if bb interesect
		if ((*i)->isAlive() && (*j)->isAlive() &&
			(*i)->isSolid() && (*j)->isSolid())
		{
			int x1 = (int)(*i)->getx();
			int y1 = (int)(*i)->gety();
			int z1 = (int)(*i)->getz();
			int dx1 = (*i)->sizex;
			int dy1 = (*i)->sizey;
			int dz1 = (*i)->sizez;
			int x2 = (int)(*j)->getx();
			int y2 = (int)(*j)->gety();
			int z2 = (int)(*j)->getz();
			int dx2 = (*j)->sizex;
			int dy2 = (*j)->sizey;
			int dz2 = (*j)->sizez;
			if(!(
					(x1 >= x2+dx2) || (x2 >= x1+dx1) || 
					(y1 >= y2+dy2) || (y2 >= y1+dy1) ||
					(z1 >= z2+dz2) || (z2 >= z1+dz1) 
				)
				)
			{
				(*i)->handleCollission ((*j));
				(*j)->handleCollission ((*i));
			}
		}		
	}
	
	testSprites();

	// remove all that are not alive!
	sprites.remove_if (MyObjectRemover());
}

static bool sortFunc (Sprite *a, Sprite *b)
{
	float rza = a->getx() + a->gety() + a->getz();
	float rzb = b->getx() + b->gety() + b->getz();
	return rza < rzb;
}

void SpriteLayer::draw (BITMAP *buffer)
{
	testSprites();
	
	int xofst = 0;
	int yofst = 0;
	if (viewPort)
	{
		xofst = viewPort->getXofst();
		yofst = viewPort->getYofst();
	}
	
	// sort sprites by z order
	sprites.sort (&sortFunc);
	
	list<Sprite*>::iterator i;
	
	for (i = sprites.begin(); i != sprites.end(); i++)
	{
		if ((*i)->isVisible() && (*i)->isAlive())
		{
			// first draw shadow if this object is floating
			if ((*i)->hasShadow())
			{
				(*i)->drawShadow(buffer, xofst, yofst);
			}
			(*i)->draw(buffer, xofst, yofst);
#ifdef DEBUG
			if (parent.isDebug())
			{
				int rx, ry;
				isoToScreen ((*i)->getx(), (*i)->gety(), (*i)->getz(), rx, ry);
				
				drawWireFrame(buffer, rx - xofst, ry - yofst, (*i)->sizex, (*i)->sizey, (*i)->sizez, CYAN);
			}
#endif
		}
	}	
}

void Sprite::drawShadow (BITMAP * buffer, int xofst, int yofst)
{
	int rx, ry;
	isoToScreen (x, y, 0, rx, ry);
	
	drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);
	int ssize = (int)(sizex * pow (0.995, z));
	set_trans_blender (0, 0, 0, 64);
	
	ellipsefill (buffer, rx - xofst, ry - yofst + (ssize / 2), 
		ssize, ssize / 2, BLACK);
	drawing_mode (DRAW_MODE_SOLID, NULL, 0, 0);	
}

void Sprite::draw(BITMAP * buffer, int xofst, int yofst)
{
	int rx, ry;
	isoToScreen (x, y, z, rx, ry);
	
	ellipsefill (buffer, rx - xofst, ry - yofst, sizex, sizex / 2, color);
	ellipse (buffer, rx - xofst, ry - yofst, sizex, sizex / 2, BLACK);
	float radians = radiansFromDir(dir);
	float dx = sizex * cos (radians);
	float dy = sizey * sin (radians);
	int rdx, rdy;
	isoToScreen (dx, dy, 0, rdx, rdy);
	line (buffer, rx - xofst, ry - yofst, rx - xofst + rdx, ry - yofst + rdy, BLACK);
}

void Sprite::try_move(float dx, float dy, float dz)
{
	x += dx;
	y += dy;
	z += dz;
	if (z < 0) z = 0;
}

/*
Sprite * SpriteLayer::getNearestSprite(SpriteType st, float x, float y)
{
	list<Sprite*>::iterator i;
	
	float minDist;
	Sprite *min = NULL;
	
	// first draw shadows
	for (i = sprites.begin(); i != sprites.end(); i++)
	{
		if ((*i)->isAlive() && (*i)->getType() == st)
		{
			float xx = (*i)->getx();
			float yy = (*i)->gety();
			float dist = (x - xx) * (x - xx) + (y - yy) * (y - yy);
			if (min == NULL || dist < minDist)
			{
				min = *i;
				minDist = dist;
			}
		}
	}
	return min;
}
*/
Sprite * SpriteLayer::getNearestSprite_ex (
	SpriteType st, 
	bool selectSubType,
	SpriteSubType sst, 
	float x, float y,
	bool selectRange,
	float maxRange, 
	bool selectSpread,
	float range_angle,
	float range_spread)
{
	list<Sprite*>::iterator i;
	
	float minDist;
	Sprite *min = NULL;
	
	// first draw shadows
	for (i = sprites.begin(); i != sprites.end(); i++)
	{
		if ((*i)->isAlive() && (*i)->getType() == st)
		{	
			bool ok = true;
			float dist, dx, dy;
			
			if (selectSubType)
			{
				ActiveSprite *a = dynamic_cast<ActiveSprite*>(*i);
				ok = (a && a->getSubType() == sst);
			}
			if (ok)
			{
				dx = x - (*i)->getx();
				dy = y - (*i)->gety();
				dist = dx * dx + dy * dy;
			}
			if (ok && selectRange)
			{
				if (dist > (maxRange * maxRange))
				{
					ok = false;
				}
			}
			if (ok && selectSpread)
			{
				float vector_angle = atan2 (dx, dy);
				float delta_angle = vector_angle - range_angle;
				normalize_angle (delta_angle);
				if (fabs(delta_angle) > range_spread)
				{
					ok = false;
				}
			}
			if (ok && (min == NULL || dist < minDist))
			{
				min = *i;
				minDist = dist;
			}
		}
	}
	return min;
}

Sprite *SpriteLayer::getNextBunker ()
{
	return 
		getNearestSprite_ex (
			ST_ENEMY, 
			true,
			SST_MAINBUNKER,
			0, 0,
			false,
			0,
			false,
			0,
			0); 
}

float radiansFromDir (Direction dir)
{
	return dir * M_PI / 4;
}

Direction dirFromRadians (float radians)
{
	int result = int(round(radians * 4 / M_PI)) & 7;
	return (Direction) result;
}

void SpriteLayer::testSprites()
{
	list<Sprite*>::iterator i;
	for (i = sprites.begin(); i != sprites.end(); i++)
	{
		assert ((*i)->getDir() < 8 && (*i)->getDir() >= 0);
		assert ((*i)->getx() != NAN);
		assert ((*i)->gety() != NAN);
	}
}
